#include "matrix.h"
#include "point.h"

Matrix::Matrix()
{
  SetIdentity();
}

Matrix::~Matrix()
{
}

VOID Matrix::SetNull()
{
  memset(m_m, 0, sizeof(MATRIX4D));
}

VOID Matrix::SetIdentity()
{
  UINT_32 i;

  SetNull();

  for ( i = 0; i < 16; i += 5 )
    {
      m_m[i] = 1.0;
    }
}

////////////////////////////////////////////////////////
// START: ASSIGNMENT 3: Fill in the 2D Transformations
// 

VOID Matrix::SetTranslate(const Vector &v)
{
  SetIdentity();
}

VOID Matrix::SetRotateZ(DOUBLE dTheta)
{
  SetIdentity();
}

VOID Matrix::SetScale(const Vector &v)
{
  SetIdentity();
}

VOID Matrix::SetScale(const FLOAT f)
{
  SetIdentity();
}

VOID Matrix::SetViewPort(const Vector &vVPMin, const Vector &vVPMax, 
			 const Vector &vWWMin, const Vector &vWWMax)
{
}

// 
// END  : ASSIGNMENT 3: Fill in the 2D Transformations
////////////////////////////////////////////////////////

FLOAT Matrix::Trace() const
{
  return ( m_m[0] + m_m[5] + m_m[10] );
}

VOID Matrix::Transpose()
{
  UINT_32 i, j;
  FLOAT f;

  for ( i = 0; i < (4-1); i++ )
    {
      for ( j = i+1; j < 4; j++)
	{
	  f = m_m[Index(i,j)];
	  m_m[Index(i,j)] = m_m[Index(j,i)];
	  m_m[Index(j,i)] = f;
	}
    }
}

VOID Matrix::Invert()
{
}

Matrix &Matrix::operator=(const Matrix& m)
{
  memcpy(m_m, m.m_m, sizeof(MATRIX4D));

  return (*this);
}

Matrix &Matrix::operator+=(const Matrix& m)
{
  UINT_32 i;

  for ( i = 0; i < 16; i++ )
    {
      m_m[i] = m_m[i] + m.m_m[i];
    }

  return (*this);
}

Matrix &Matrix::operator-=(const Matrix& m)
{
  UINT_32 i;

  for ( i = 0; i < 16; i++ )
    {
      m_m[i] = m_m[i] - m.m_m[i];
    }

  return (*this);
}

Matrix &Matrix::operator*=(const FLOAT f)
{
  UINT_32 i;

  for ( i = 0; i < 16; i++ )
    {
      m_m[i] = m_m[i] * f;
    }

  return (*this);
}

Matrix &Matrix::operator/=(const FLOAT f)
{
  UINT_32 i;

  ASSERT( f != 0.0 );

  for ( i = 0; i < 16; i++ )
    {
      m_m[i] = m_m[i] / f;
    }

  return (*this);
}

Matrix Matrix::operator*(const Matrix& m) const
{
  UINT_32 i, j, k;
  UINT_32 i4, k4;

  Matrix mResult;

  i4 = 0;
  for ( i = 0; i < 4; i++ ) // Col Selector
    {
      for ( j = 0; j < 4; j++ ) // Row Selector
	{
	  mResult.m_m[i4+j] = 0.0;
	  k4 = 0;
	  for ( k = 0; k < 4; k++ ) // Iterator
	    {
	      mResult.m_m[i4+j] += m_m[k4+j] * m.m_m[i4+k];
	      k4 += 4;
	    }
	}
      i4 += 4;
    }

  return mResult;
}

Vector Matrix::operator*(const Vector &v) const
{
  UINT_32 i, j, j4;
  Vector vResult;

  for ( i = 0; i < 4; i++ )
    {
      vResult[i] = 0.0f;
      j4 = 0;
      for ( j = 0; j < 4; j++ )
	{
	  vResult[i] += m_m[i + j4] * v(j);
	  j4 += 4;
	}
    }

  return vResult;
}

Vector operator*(const Vector &v, const Matrix &m)
{
  UINT_32 i, j, k;
  Vector vResult;
  MATRIX4D mTemp;

  m.Get(mTemp);

  k = 0;
  for ( i = 0; i < 4; i++ )
    {
      vResult[i] = 0.0f;
      for ( j = 0; j < 4; j++ )
	{
	  vResult[i] += mTemp[k] * v(j);
	  k++;
	}
    }

  return vResult;
}

Point Matrix::operator*(const Point &pt) const
{
  UINT_32 i, j, j4;
  Point ptResult;

  for ( i = 0; i < 4; i++ )
    {
      ptResult[i] = 0.0f;
      j4 = 0;
      for ( j = 0; j < 4; j++ )
	{
	  ptResult[i] += m_m[i + j4] * pt(j);
	  j4 += 4;
	}
    }

  return ptResult;
}

VOID Matrix::Display() const
{
  UINT_32 i, j;

  for ( i = 0; i < 4; i++ )
    {
      printf("| ");
      for ( j = 0; j < 4; j++ )
	{
	  printf("%1.3f ", m_m[Index(i,j)]);
	}
      printf("|\n");
    }
}




