#include "slidegroup.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEGroup Class
//

CSLIDEGroup::CSLIDEGroup()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;
}

CSLIDEGroup::~CSLIDEGroup()
{
}

VOID CSLIDEGroup::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID)
{
  CSLIDENode::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;
  m_dyniShading.SetValue(slideshtSLF_INHERIT);
  m_dyniLOD.SetValue(slidelodtSLF_FULL);
  m_listCSLIDEInstance.Init();
}

VOID CSLIDEGroup::Uninit()
{
  CSLIDENode::Uninit();
}

BOOL CSLIDEGroup::Initialized()
{
  if ( !CSLIDENode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEGroup::InsertInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.InsertLast(pCSLIDEInstance, 0);
}

BOOL CSLIDEGroup::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDENode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface], 
		  m_pcSurfaceID);
      return FALSE;
    }

  ASSERT( 0 <= m_dyniShading.GetValue() &&
	  m_dyniShading.GetValue() < slideshtSize );
  ASSERT( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slidelodtSize );

  return TRUE;
}

BOOL CSLIDEGroup::Valid()
{
  if ( !CSLIDENode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDEPrimitiveIter *CSLIDEGroup::AllocateIter()
{
  CSLIDEGroupIter *pCSLIDEGroupIter;

  pCSLIDEGroupIter = new CSLIDEGroupIter;
  ASSERT(pCSLIDEGroupIter != NULL );
  pCSLIDEGroupIter->Init(this);

  return pCSLIDEGroupIter;
}

VOID CSLIDEGroup::FreeIter(CSLIDEPrimitiveIter *pCSLIDEGroupIter)
{
  delete pCSLIDEGroupIter;
}

CSLIDECamera *CSLIDEGroup::FindCamera(CListIterInt *piterIDs, 
				      CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );
  
  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      HandleError("Bad %s path: path ends in %s \"%s\"",
		  gcppcSLIDEReservedWords[sliderwtCamera],
		  gcppcSLIDEReservedWords[sliderwtGroup], GetID());
      return NULL;
    }
  
  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if ( pCSLIDEInstance->GetID() == pcID )
	{
	  plistCSLIDEInstance->InsertLast(pCSLIDEInstance, 0);
	  return pCSLIDEInstance->FindCamera(piterIDs, plistCSLIDEInstance);
	}
    }
  
  HandleError("Bad %s path: %s \"%s\" does not contain an %s \"%s\"",
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtGroup], GetID(),
	      gcppcSLIDEReservedWords[sliderwtInstance], pcID);
  return NULL;
}

CSLIDELight *CSLIDEGroup::FindLight(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      HandleError("Bad %s path: path ends in %s \"%s\"",
		  gcppcSLIDEReservedWords[sliderwtLight],
		  gcppcSLIDEReservedWords[sliderwtGroup], GetID());
      return NULL;
    }

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if ( pCSLIDEInstance->GetID() == pcID )
	{
	  plistCSLIDEInstance->InsertLast(pCSLIDEInstance, 0);
	  return pCSLIDEInstance->FindLight(piterIDs, plistCSLIDEInstance);
	}
    }

  HandleError("Bad %s path: %s \"%s\" does not contain an %s \"%s\"",
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtGroup], GetID(),
	      gcppcSLIDEReservedWords[sliderwtInstance], pcID);
  return NULL;
}

////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 3: Fill these in
// 
VOID CSLIDEGroup::Preprocess()
{
}

VOID CSLIDEGroup::CheckDynamic()
{
}

VOID CSLIDEGroup::PreprocessGraph()
{
}

VOID CSLIDEGroup::Update()
{
}

VOID CSLIDEGroup::UpdateState()
{
}

VOID CSLIDEGroup::UpdateGraph()
{
}

VOID CSLIDEGroup::InsertIntoBound(const Matrix &m, CBound &bound)
{
}

VOID CSLIDEGroup::CalculateBound()
{
}

VOID CSLIDEGroup::ModifySurfacePointer(CSLIDESurface* &pCSLIDESurface)
{
}
// 
// END  : ASSIGNMENT 3: Fill these in
////////////////////////////////////////////////////////////////////////

VOID CSLIDEGroup::ModifyShadingFlag(SLIDEShadingType &/*eShading*/)
{
  // Ignore shading flags
}

VOID CSLIDEGroup::ModifyLODFlag(SLIDELODType &/*eLOD*/)
{
  // Ignore LOD flags
}

////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 3: Fill these in
// 
VOID CSLIDEGroup::RenderGeometry(const Matrix &mGvp_proj, 
				 const Matrix &mGproj_vrc, 
				 const Matrix &mGvrc_proj, 
				 const Matrix &mGvrc_obj, 
				 const Matrix &mGobj_vrc, 
				 CBound *pCBoundClip, 
				 CListInt *plistLights,
				 CSLIDESurface *pCSLIDESurface,
				 SLIDEShadingType eShading, 
				 SLIDELODType eLOD)
{
}

VOID CSLIDEGroup::RenderBound(const Matrix &mGvp_proj, const Matrix &mGproj_vrc, 
			      const Matrix &mGvrc_obj)
{
  gilPushAttrib(GL_CURRENT_BIT);
  {
    if ( m_bBoundDynamic && GetColorDynamicBounds() )
      {
	// Red for Dynamic Bounding Boxes
	gilColor3f(1.0, 0.0, 0.0);
      }
    else
      {
        // White for Static Bounding Boxes
	gilColor3f(1.0, 1.0, 1.0);
      }
    m_bound.Render(mGvp_proj, mGproj_vrc * mGvrc_obj);
  }
  gilPopAttrib();
}
// 
// END  : ASSIGNMENT 3: Fill these in
////////////////////////////////////////////////////////////////////////

VOID CSLIDEGroup::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtGroup], m_pcID);

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  if ( m_dyniShading.GetValue() != slideshtSLF_INHERIT )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtShading], 
	      gcppcSLIDEShadingReservedWords[m_dyniShading.GetValue()]);
    }
  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndGroup]);
}

UINT_32 CSLIDEGroup::GetInstanceCount()
{
  return m_listCSLIDEInstance.GetLength();
}

//////////////////////////////////////////////////////////////////////////////
// CSLIDEGroupIter Class
//

CSLIDEGroupIter::CSLIDEGroupIter()
{
}

CSLIDEGroupIter::~CSLIDEGroupIter()
{
}

VOID CSLIDEGroupIter::Init(CSLIDEGroup *pCSLIDEGroup)
{
  CSLIDEPrimitiveIter::Init(pCSLIDEGroup);

  m_iter.Init(&(pCSLIDEGroup->m_listCSLIDEInstance));
}

VOID CSLIDEGroupIter::Uninit()
{
  CSLIDEPrimitiveIter::Uninit();
}

CSLIDEInstance *CSLIDEGroupIter::PeekFirstCSLIDEInstance()
{
  return (CSLIDEInstance*)(m_iter.PeekFirst());
}

CSLIDEInstance *CSLIDEGroupIter::PeekNextCSLIDEInstance()
{
  return (CSLIDEInstance*)(m_iter.PeekNext());
}
