#ifndef __SLIDESURFACE_H__
#define __SLIDESURFACE_H__

#include "slideprimitive.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDESurface Class
//

class CSLIDESurface : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESurface();
  virtual ~CSLIDESurface();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:
////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 3: These are the methods you will be dealing with
// 
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();


//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Push();   // Use these
  virtual VOID Pop();
  virtual VOID Render(); // Do not use this directly
// 
// END  : ASSIGNMENT 3: These are the methods you will be dealing with
////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID SetColor(const Vector &vColor);
  inline VOID SetColor(DynVector *pdynvColor);
  inline const Vector &GetColor() const;
  inline VOID SetReflectivity(const Vector &vReflectivity);
  inline VOID SetReflectivity(DynVector *pdynvReflectivity);
  inline const Vector &GetReflectivity() const;
  inline VOID SetExponent(FLOAT fExponent);
  inline VOID SetExponent(DynFloat *pdynfExponent);
  inline FLOAT GetExponent() const;
  inline VOID SetMetallic(FLOAT fMetallic);
  inline VOID SetMetallic(DynFloat *pdynfMetallic);
  inline FLOAT GetMetallic() const;

  static inline CSLIDESurface *GetSLF_INHERIT();

//////////////////////////////////////////////////
// Class Variables
//
public:
  static CSLIDESurface *m_pCSLIDESurfaceDefault;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynVector m_vColor;         // Color Vector RGB
  DynVector m_vReflectivity;  // Ignore this
  DynFloat m_dynfExponent;    // Ignore this
  DynFloat m_dynfMetallic;    // Ignore this
};

//////////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDESurface::SetColor(const Vector &vColor)
{
  m_vColor = vColor;
}

inline VOID CSLIDESurface::SetColor(DynVector *pdynvColor)
{
  m_vColor.SetAll(pdynvColor);
}

inline const Vector &CSLIDESurface::GetColor() const
{
  return m_vColor;
}

inline VOID CSLIDESurface::SetReflectivity(const Vector &vReflectivity)
{
  m_vReflectivity = vReflectivity;
}

inline VOID CSLIDESurface::SetReflectivity(DynVector *pdynvReflectivity)
{
  m_vReflectivity.SetAll(pdynvReflectivity);
}

inline const Vector &CSLIDESurface::GetReflectivity() const
{
  return m_vReflectivity;
}

inline VOID CSLIDESurface::SetExponent(FLOAT fExponent)
{
  m_dynfExponent.SetValue(fExponent);
}

inline VOID CSLIDESurface::SetExponent(DynFloat *pdynfExponent)
{
  m_dynfExponent.Set(pdynfExponent);
}

inline FLOAT CSLIDESurface::GetExponent() const
{
  return m_dynfExponent.GetValue();
}

inline VOID CSLIDESurface::SetMetallic(FLOAT fMetallic)
{
  m_dynfMetallic.SetValue(fMetallic);
}

inline VOID CSLIDESurface::SetMetallic(DynFloat *pdynfMetallic)
{
  m_dynfMetallic.Set(pdynfMetallic);
}

inline FLOAT CSLIDESurface::GetMetallic() const
{
  return m_dynfMetallic.GetValue();
}

inline CSLIDESurface *CSLIDESurface::GetSLF_INHERIT()
{
  return m_pCSLIDESurfaceDefault;
}

#endif // __SLIDESURFACE_H__


