#include "window.h"

#include "slideparser.h"
#include "slideworld.h"
#include "slideface.h"
#include "slidenode.h"
#include "slideviewport.h"
#include "sliderender.h"

CWindow *gpCWindow = NULL;
CWindow *gpCWindowCurr = NULL;

static CHAR gpcBuffer[200];

static const INT_32 ciWindowSize = 500;

//////////////////////////////////////////////////////////////////////
// CWindow Class
//

Tcl_Interp *CWindow::m_pInterp = NULL;

CHashString *CWindow::m_pHashKeyboard = NULL;

const int CWindow::m_cuiOotMethods = cuiCWindowOotMethods;

// OOT Registration Variables
OotMethodEntry CWindow::m_pOotMethods[cuiCWindowOotMethods] =
{
  CWINDOW_METHODS
};
const CHAR *CWindow::m_cppcOotFields[CWindow::ftSize] =
{
  CWINDOW_FIELDS_STRINGS
};
const CHAR *CWindow::m_cpcOotClassName = "window";
OotClass *CWindow::m_pOotClass = NULL;

UINT_32 CWindow::m_uiOotCounter = 0;

CHAR *CWindow::m_pcUIConfigFile = NULL;

// Local Defaults
//static const BOOL cbDisplayEvents = TRUE;
static const BOOL cbDisplayEvents = FALSE;
static const CWindow::EventLevelType ceEventLevel = CWindow::eltAction;
//static const CWindow::EventLevelType ceEventLevel = CWindow::eltCallBack;

CWindow::CWindow()
{
  m_uiOotID = 0;
  m_pcOotName = NULL;
  m_pOotObject = NULL;

  m_pTogl = FALSE;
  m_bRedisplayPending = FALSE;

  m_pcFile = NULL;
  m_uiLine = 0;
  m_uiColumn = 0;
  m_pcSLIDEID = NULL;
  m_pCSLIDEWorld = NULL;
}

CWindow::~CWindow()
{
  Uninit();
}

VOID CWindow::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		   CHAR *pcSLIDEID)
{
  UINT_32 i;

  ASSERT( pcFile != NULL );
  ASSERT( pcSLIDEID != NULL );

  // Set Up Keyboard Interpreter
  // This is only done once for the whole CWindow Class
  if ( m_pHashKeyboard == NULL )
    {
      m_pHashKeyboard = new CHashString;
      ASSERT( m_pHashKeyboard != NULL );
      m_pHashKeyboard->Init((3*gcuiglsiKeyBindings)/2);
      
      for ( i = 0; i < gcuiglsiKeyBindings; i++ )
	{
	  m_pHashKeyboard->Insert(&(gcpglsiKeyBindings[i]), gcpglsiKeyBindings[i].pcKey);
	}
    }

  m_uiOotID = 0;
  m_pcOotName = NULL;
  m_pOotObject = NULL;

  m_pTogl = FALSE;
  m_bRedisplayPending = FALSE;

  m_iWidth = ciWindowSize;
  m_iHeight = ciWindowSize;

  m_bDisplayEvents = cbDisplayEvents;
  m_eEventLevel = ceEventLevel;

  // SLIDE Stuff
  m_pcFile = pcFile;
  m_uiLine = uiLine;
  m_uiColumn = uiColumn;
  m_pcSLIDEID = pcSLIDEID;
  m_pCSLIDEWorld = NULL;
  m_listCSLIDEViewport.Init();

  m_vOrigin.Null();
  m_vSize.Set(1.0, 1.0, 0.0);
  m_vBackground.Null();

  // Output Stuff
  m_pcOutputFile = NULL;
  m_fpOutputFile = NULL;

  // Set up the OOT communication link with Tcl
  InitTcl();

  // Create the actual Togl drawing area and initialize OpenGL
  //InitTogl();

  SetDrawBounds(FALSE);
  SetColorDynamicBounds(FALSE);
  SetDrawViewportBound(FALSE);

  SetSecondsPerFrameRequested(CSLIDEWorld::GetSecondsPerFrameRequested());
  SetUpdateTime(0.0);
  SetRenderTime(0.0);  
}

VOID CWindow::InitTcl()
{
  //////////////////////////////////////////////////
  // Tcl Stuff
  m_uiOotID = (++m_uiOotCounter);
//  sprintf(gpcBuffer, "%s%ld", m_cpcOotClassName, m_uiOotID);
  sprintf(gpcBuffer, "slf%s", m_pcSLIDEID);
  m_pcOotName = new CHAR[strlen(gpcBuffer) + 1];
  ASSERT( m_pcOotName != NULL );
  strcpy(m_pcOotName, gpcBuffer);

  m_pOotObject = Oot_CreateObject(m_pInterp, m_pOotClass, m_pcOotName, this);
  ASSERT( m_pOotObject != NULL );
}

VOID CWindow::InitTogl()
{
  INT_32 iCode;

  ASSERT( gpCWindowCurr == NULL );
  gpCWindowCurr = this;
/*
  // Create Togl Drawing Area
  sprintf(gpcBuffer, "frame .%s; togl .%s.toglarea -width 400 -height 400 -rgba true -double true -depth true -stencil true -ident %s", m_pcOotName, m_pcOotName, m_pcOotName);
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  */
  //printf("Created Togl Area\n");

  //////////////////////////////////////////////////
  // Window UI Config File
  if ( GetUIConfigFile() == NULL )
#ifdef WIN32
    {
      HKEY hkey, hkeyResult;
      LONG lResult;
      DWORD dwSize;
      ULONG ulType;
      
      lResult = RegOpenKeyEx(HKEY_CURRENT_USER, "Software", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slideview2d\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      hkey = hkeyResult;
      
      lResult = RegOpenKeyEx(hkey, "UC Berkeley", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slideview2d\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;
      
      lResult = RegOpenKeyEx(hkey, "SLIDE", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slideview2d\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;

      lResult = RegOpenKeyEx(hkey, "1.0", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slideview2d\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;

      lResult = RegOpenKeyEx(hkey, "slideview2d", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slideview2d\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;

      dwSize = 200;
      lResult = RegQueryValueEx(hkeyResult, "UIConfigFile", 0, &ulType, (LPBYTE)gpcBuffer, &dwSize);
      if ( lResult != ERROR_SUCCESS || ulType != REG_SZ )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slideview2d\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}

      CWindow::SetUIConfigFile(gpcBuffer);
    }
#else // WIN32
    {
      SetUIConfigFile("/project/cs/sequin/caffe/slide/lib/slideview2d/window.tcl");
    }
#endif // WIN32

  ASSERT( m_pcUIConfigFile != NULL );
  iCode = Tcl_EvalFile(m_pInterp, m_pcUIConfigFile); 
  if ( iCode != TCL_OK )
    {
      sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error opening UI configure file: \\\"%s\\\"\\n Exiting program.\"", m_pcUIConfigFile);
      iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
      if ( iCode != TCL_OK )
	{
	  printf("MessageBox returned: \"%s\"\n", m_pInterp->result);
	}
      ASSERT( iCode == TCL_OK );
      exit(1);
    }

  //printf("Sourced CreateWindow Function\n");

  sprintf(gpcBuffer, "CreateWindow . %s", m_pcOotName);
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  //printf("Created Window\n");
  /*
  sprintf(gpcBuffer, "pack .%s", m_pcOotName);
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  */
  //printf("Packed Window\n");
}

VOID CWindow::HandleDestroy()
{
  INT_32 iCode;

  if ( m_pInterp != NULL && m_pOotClass != NULL && m_pcOotName != NULL )
    {
      sprintf(gpcBuffer, "DestroyWindow . %s", m_pcOotName);
      iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );

      Oot_DestroyObject(m_pInterp, m_pOotClass, m_pcOotName);
    }
}

VOID CWindow::Uninit()
{
  if ( m_pcOotName != NULL )
    {
      delete []m_pcOotName;
      m_pcOotName = NULL;
    }

  // We do not own this memory
  m_pOotObject = NULL;
  m_pTogl = NULL;
  m_pCSLIDEWorld = NULL;

  // Output File
  if ( m_pcOutputFile != NULL )
    {
      delete []m_pcOutputFile;
      m_pcOutputFile = NULL;
    }
  ASSERT( m_fpOutputFile == NULL );
}

VOID CWindow::InsertViewport(CSLIDEViewport *pCSLIDEViewport)
{
  ASSERT( pCSLIDEViewport != NULL );

  m_listCSLIDEViewport.InsertLast(pCSLIDEViewport, 0);
}

BOOL CWindow::CreateTclToplevel()
{
  INT_32 iCode;

  // Create Togl Drawing Area
  sprintf(gpcBuffer, "toplevel .%s", m_pcOotName);
  iCode = Tcl_GlobalEval(m_pInterp, gpcBuffer); 
  return ( iCode == TCL_OK );
}

BOOL CWindow::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  m_pCSLIDEWorld = pCSLIDEWorld;

  return TRUE;
}

VOID CWindow::Preprocess()
{
  // Open the Window
  // Create the actual Togl drawing area and initialize OpenGL
  InitTogl();
}

VOID CWindow::Update()
{
  m_vBackground.Update();

  PostRedisplay();
}

VOID CWindow::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtWindow], m_pcSLIDEID);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtOrigin], 
	  m_vOrigin[X], m_vOrigin[Y]);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtWinSize], 
	  m_vSize[X], m_vSize[Y]);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtBackground], 
	  m_vBackground[X], m_vBackground[Y], m_vBackground[Z]);
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndWindow]);
}

VOID CWindow::HandleReshape(INT_32 iWidth, INT_32 iHeight)
{
  DisplayEvent(eltCallBack, "Entering HandleReshape(%3ld, %3ld)\n", 
	       iWidth, iHeight);

  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  m_iWidth = iWidth;
  m_iHeight = iHeight;

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->HandleReshape((UINT_32)m_iWidth, (UINT_32)m_iHeight);
    }  

  gilViewport(0, 0, m_iWidth, m_iHeight);

  gilMatrixMode(GL_PROJECTION);
  gilLoadIdentity();
  //giluOrtho2D(0, 1, 0, 1);
  //giluOrtho2D(0, m_iWidth, 0, m_iHeight);
  gilOrtho(0, m_iWidth, 0, m_iHeight, -1000, 1000);

  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();
}

VOID CWindow::HandleDisplay()
{
  DisplayEvent(eltCallBack, "Entering HandleDisplay()\n");

  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;
  Timer t;

  t.Reset();
  t.Start();

  m_bRedisplayPending = FALSE;

  // Make our Togl Window the current OpenGL Context
  ASSERT( m_pTogl != NULL );
  Togl_MakeCurrent(m_pTogl);

  gilClearColor(m_vBackground[0], m_vBackground[1], m_vBackground[2], 
		m_vBackground[3]);
  gilClear(GL_COLOR_BUFFER_BIT );
  gilDisable(GL_DEPTH_TEST);
  gilShadeModel(GL_SMOOTH);
  gilDisable(GL_CULL_FACE);
  gilDisable(GL_TEXTURE_2D);
  gilDepthFunc(GL_LEQUAL);

  gilDisable(GL_LIGHTING);
  gilDisable(GL_NORMALIZE);

  gilColor3f(0.0, 1.0, 1.0);

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->Render(0, 0, 
			      (UINT_32)m_iWidth, (UINT_32)m_iHeight);
    }

  Togl_SwapBuffers(m_pTogl);

  t.Stop();
  SetUpdateTime(CSLIDEWorld::GetUpdateTime());
  SetRenderTime(t.GetSeconds());
}

VOID CWindow::HandleMouseButton(INT_32 iButton, INT_32 iState, INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleMouseButton(%ld, %ld, %3ld, %3ld)\n", 
	       iButton, iState, iX, iY);

  BOOL b, bPostRedisplay;
  FLOAT fX, fY;
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  if ( iState == GLSI_DOWN )
    {
      // Fill in the appropriate actions and remember to set
      // bPostRedisplay = TRUE if you want to redraw any changes
      // to your data structures.

      switch( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  DisplayEvent(eltAction, "L down  [%3ld %3ld]\n", iX, iY);
	  break;
	  
	case GLSI_MIDDLE_BUTTON:
	  DisplayEvent(eltAction, "M down  [%3ld %3ld]\n", iX, iY);
	  break;
	
	case GLSI_RIGHT_BUTTON:
	  DisplayEvent(eltAction, "R down  [%3ld %3ld]\n", iX, iY);
	  break;
	}
    }
  else if ( iState == GLSI_UP )
    {
      switch( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  DisplayEvent(eltAction, "L up    [%3ld %3ld]\n", iX, iY);
	  break;
	  
	case GLSI_MIDDLE_BUTTON:
	  DisplayEvent(eltAction, "M up    [%3ld %3ld]\n", iX, iY);
	  break;
	
	case GLSI_RIGHT_BUTTON:
	  DisplayEvent(eltAction, "R up    [%3ld %3ld]\n", iX, iY);
	  break;
	}
    }

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->HandleMouseButton((UINT_32)m_iWidth, 
					     (UINT_32)m_iHeight,
					     iButton, iState, 
					     iX - 0, iY - 0);
      bPostRedisplay = bPostRedisplay || b;
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleMouseMotion(INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleMouseMotion(%3ld, %3ld)\n", 
	       iX, iY);

  BOOL b, bPostRedisplay;
  FLOAT fX, fY;
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  //sprintf(gpcBuffer, "%f", MIN(1.0, MAX(fX, 0.0)));
  //Tcl_SetVar(m_pInterp, "SLF_MOUSE_X", gpcBuffer, TCL_GLOBAL_ONLY);
  //sprintf(gpcBuffer, "%f", MIN(1.0, MAX(fY, 0.0)));
  //Tcl_SetVar(m_pInterp, "SLF_MOUSE_Y", gpcBuffer, TCL_GLOBAL_ONLY);

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->HandleMouseMotion((UINT_32)m_iWidth, 
					     (UINT_32)m_iHeight,
					     iX - 0, iY - 0);
      bPostRedisplay = bPostRedisplay || b;
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleKeyboard(INT_32 iKey, INT_32 iState, INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleKeyboard('%c', %3ld, %3ld)\n",
	       (CHAR)iKey, iX, iY);

  BOOL bPostRedisplay;
  FLOAT fX, fY;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  if ( iState != GLSI_DOWN )
    {
      return;
    }

  switch ( iKey )
    {
    case GLSI_KEY_ESCAPE:  // Escape Key
      exit(0);
      break;

    case 'b':
      SetDrawBounds(!CSLIDENode::GetRenderBound());
      bPostRedisplay = TRUE;
      break;

    case 'd':
      SetColorDynamicBounds(!CSLIDENode::GetColorDynamicBounds());
      bPostRedisplay = TRUE;
      break;

    case 'v':
      SetDrawViewportBound(!CSLIDEViewport::GetRenderViewportBound());
      bPostRedisplay = TRUE;
      break;

    case 'w':
      HandleWrite();
      break;

    case 's':
      HandleStartDynamics();
      break;

    case ' ':
      HandlePauseDynamics();
      break;

    case 'j':
      printf("CS184 Solution Code!!!\n");
      break;

    default:
      DisplayEvent(eltAction, "Key '%c' [%3ld %3ld]\n", (CHAR)iKey, iX, iY);
      break;
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleWrite()
{
  INT_32 iCode;
  UINT_32 uiLength;

  sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .slf -filetypes {{\"SLIDE Files\" {.slf}}}");
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  uiLength = strlen(m_pInterp->result);
  if ( uiLength == 0 )
    {
      // The user hit Cancel
      return;
    }

  if ( m_pcOutputFile != NULL )
    {
      delete []m_pcOutputFile;
      m_pcOutputFile = NULL;
    }
  ASSERT( m_fpOutputFile == NULL );

  ASSERT( m_pInterp->result != NULL );
  m_pcOutputFile = new CHAR[uiLength + 1];
  ASSERT( m_pcOutputFile != NULL );
  strcpy(m_pcOutputFile, m_pInterp->result);

  while ( m_fpOutputFile == NULL )
    {
      m_fpOutputFile = fopen(m_pcOutputFile, "w");
      if ( m_fpOutputFile == NULL )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", m_pcOutputFile);
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  
	  if ( strcmp(m_pInterp->result, "cancel") == 0 )
	    {
	      // Cancel
	      return;
	    }
	}
    }

  WriteSLIDE(m_fpOutputFile);

  if ( m_fpOutputFile != stdout )
    {
      ASSERT( m_fpOutputFile != NULL );
      fclose(m_fpOutputFile);
    }

  m_fpOutputFile = NULL;
}

VOID CWindow::HandleStartDynamics()
{
  m_pCSLIDEWorld->RestartDynamics();  
}

VOID CWindow::HandlePauseDynamics()
{
  m_pCSLIDEWorld->PauseDynamics();  
}

VOID CWindow::SetDrawBounds(BOOL b)
{
  CSLIDENode::SetRenderBound(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawBounds], b);
}

VOID CWindow::SetColorDynamicBounds(BOOL b)
{
  CSLIDENode::SetColorDynamicBounds(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftColorDynamicBounds], b);
}

VOID CWindow::SetDrawViewportBound(BOOL b)
{
  CSLIDEViewport::SetRenderViewportBound(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawViewportBound], b);
}

VOID CWindow::SetSecondsPerFrameRequested(DOUBLE dSecondsPerFrame)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftFramesPerSecondRequested],
		     (dSecondsPerFrame == 0.0) ? (0.0) : (1.0/dSecondsPerFrame));
}

VOID CWindow::SetUpdateTime(DOUBLE dUpdateTime)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftUpdateTime], dUpdateTime);
}

VOID CWindow::SetRenderTime(DOUBLE dRenderTime)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftRenderTime], dRenderTime);
}

VOID CWindow::PostRedisplay()
{
  if ( !m_bRedisplayPending )
    { 
      ASSERT( m_pTogl != NULL );
      Togl_PostRedisplay(m_pTogl);
      m_bRedisplayPending = TRUE;
    }
}

VOID CWindow::DisplayControls()
{
}

VOID CWindow::DisplayEvent(EventLevelType eEventLevel, CHAR *pcFormat, ...)
{
  if ( m_bDisplayEvents && eEventLevel <= m_eEventLevel )
    {
      va_list vaArgs;

      va_start(vaArgs, pcFormat);
      vfprintf(stderr, pcFormat, vaArgs);
      va_end(vaArgs);
    }
}

//////////////////////////////////////////////////////////////////////
// Callback Functions
//

int CWindow::CreateOotClass(Tcl_Interp *pInterp)
{
  ASSERT( pInterp != NULL );

  m_pInterp = pInterp;

  m_pOotClass = Oot_CreateClass(m_pInterp, m_cpcOotClassName, 
				CWindow::CBHandleCreate, 
				CWindow::m_pOotMethods, 
				CWindow::m_cuiOotMethods,
				CWindow::m_cppcOotFields, 
				(UINT_32)(CWindow::ftSize));
  if ( m_pOotClass == NULL )
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}
    
int CWindow::CBHandleCreate(Tcl_Interp *pInterp, OotClass *pClass, 
			    int argc, char ** /*argv*/)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pClass != NULL );

  if ( argc != 2 )
    {
      Tcl_SetResult(pInterp, "wrong args: window <name> <world>", NULL);
      return TCL_ERROR;
    }

  pCWindow = new CWindow;
  ASSERT( pCWindow != NULL );
  pCWindow->Init(NULL, 0, 0, NULL);
  /*
  if ( !pCWindow->Init() )
    {
      delete pCWindow;
      Tcl_SetResult(pInterp, "Error initializing window", NULL);
      return TCL_ERROR;      
    }
    */
  return TCL_OK;
}

VOID CWindow::CBHandleDestroy(ClientData pvClientData)
{
  CWindow *pCWindow;

  ASSERT( pvClientData != NULL );

  pCWindow = (CWindow*)pvClientData;

  pCWindow->HandleDestroy();
}

VOID CWindow::CBHandleToglCreate(struct Togl *pTogl)
{
  ASSERT( pTogl != NULL );

  ASSERT( gpCWindowCurr != NULL );
  //printf("Togl client data = %s\n", gpCWindowCurr->GetSLIDEID());
  Togl_SetClientData(pTogl, (ClientData)gpCWindowCurr);
  gpCWindowCurr->SetTogl(pTogl);
  gpCWindowCurr = NULL;
}

VOID CWindow::CBHandleDisplay(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleDisplay();
}

VOID CWindow::CBHandleReshape(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleReshape(Togl_Width(pTogl), Togl_Height(pTogl));
}

VOID CWindow::CBHandleDestroy(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  //delete pCWindow;

  Togl_SetClientData(pTogl, NULL);
}

int CWindow::CBHandleMouseButton(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 iButton, INT_32 iState,
				 INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iNewButton, iNewState;
  INT_32 iXVP, iYVP;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  switch ( iButton )
    {
    case 1:
      iNewButton = GLSI_LEFT_BUTTON;
      break;
    case 2:
      iNewButton = GLSI_MIDDLE_BUTTON;
      break;
    case 3:
      iNewButton = GLSI_RIGHT_BUTTON;
      break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  iNewState = ( iState == 0 ) ? ( GLSI_DOWN ) : ( GLSI_UP );

  pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);

  pCWindow->HandleMouseButton(iNewButton, iNewState, iXVP, iYVP);

  return TCL_OK;
}

int CWindow::CBHandleMouseMotion(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 /*iButton*/, INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iXVP, iYVP;
  
  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);

  pCWindow->HandleMouseMotion(iXVP, iYVP);

  return TCL_OK;
}

int CWindow::CBHandleKeyboard(Tcl_Interp *pInterp, OotObject *pObject,
			      CHAR *pcKeyCode, INT_32 iState, 
			      INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iKey, iNewState;
  glsiKeyBinding *pKeyBinding;
  INT_32 iXVP, iYVP;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pKeyBinding = (glsiKeyBinding*)(m_pHashKeyboard->Peek(pcKeyCode));
  if ( pKeyBinding != NULL )
    {
      iKey = pKeyBinding->iValue;
      
      iNewState = ( iState == 0 ) ? ( GLSI_DOWN ) : ( GLSI_UP );
      pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);
      
      pCWindow->HandleKeyboard(iKey, iNewState, iXVP, iYVP);
    }

  return TCL_OK;
}

int CWindow::CBHandleWrite(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleWrite();

  return TCL_OK;
}

int CWindow::CBHandleStartDynamics(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleStartDynamics();

  return TCL_OK;
}

int CWindow::CBHandlePauseDynamics(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandlePauseDynamics();

  return TCL_OK;
}

int CWindow::CBSetDrawBounds(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawBounds(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetColorDynamicBounds(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetColorDynamicBounds(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetDrawViewportBound(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawViewportBound(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}
